"""Driver factory/creation."""

from enum import Enum
from typing import Mapping, Optional, Type

from selenium.webdriver import (
    Firefox as FirefoxDriver, FirefoxOptions,
    Chrome as ChromeDriver, ChromeOptions
)
from selenium.webdriver.remote.webdriver import WebDriver

from ..util import LoggerMixin
from .driver import Driver

class DriverType(str, Enum):
    """Common driver types."""
    CHROME = "chrome"
    FIREFOX = "firefox"

    def __str__(self):
        return self.value

class _SeleniumMapping:
    """Data structure housing Selenium driver parameters"""
    __slots__ = ("driver_type", "options_type")

    def __init__(self, driver_type: Type[WebDriver], options_type: Type):
        self.driver_type = driver_type
        self.options_type = options_type

_webdriver_type_mapping: Mapping[DriverType, _SeleniumMapping] = {
    DriverType.CHROME: _SeleniumMapping(ChromeDriver, ChromeOptions),
    DriverType.FIREFOX: _SeleniumMapping(FirefoxDriver, FirefoxOptions)
}
def _webdriver_type(name: DriverType) -> Type[WebDriver]:
    return _webdriver_type_mapping[name].driver_type

def _options_type_for(name: DriverType) -> Type:
    return _webdriver_type_mapping[name].options_type

class DriverFactory(LoggerMixin):
    def create(self, driver_type: Optional[DriverType] = None) -> Driver:
        """Construct driver with the provided driver type.

        :param driver_type: type of driver to construct.
        """
        if not driver_type:
            driver = Driver(self._try_selenium_drivers())
        else:
            driver = Driver(self._get_selenium_driver(driver_type))

        return driver

    def _get_options(self, options_type) -> object:
        options = options_type()
        options.headless = True

        return options

    def _get_selenium_driver(self, driver_type: DriverType) -> WebDriver:
        """Construct a selenium WebDriver with the provided type.

        :param driver_type: type of driver to be created.
        """
        webdriver_type = _webdriver_type(driver_type)
        options_type = _options_type_for(driver_type)

        options = options_type()
        options.headless = True

        return webdriver_type(options=options)

    def _try_selenium_drivers(self) -> WebDriver:
        """Attempt to construct any driver type."""
        exceptions = {}

        for driver_type in DriverType:
            try:
                driver = self._get_selenium_driver(driver_type)
                self.log.info(f"successfully initialized {driver_type} driver")

                return driver
            except (OSError, Exception) as e:
                exceptions[driver_type] = e

        self.log.error("unable to initialize WebDriver")
        for (driver_type, exception) in exceptions.items():
            self.log.error("%s trace:" % (driver_type))
            self.log.exception(exception)

        raise Exception("unable to initialize WebDriver")

def create(driver_type: Optional[DriverType] = None):
    """Construct driver with the provided driver type.

    :param driver_type: type of driver to construct.
    """
    return DriverFactory().create(driver_type)



"""Element wrapper/decorator."""

from typing import cast, List, Optional

from selenium.webdriver.remote.webelement import WebElement

from ..util import LoggerMixin
from .traverser import Traverser, Traversable

class ElementTraverserMixin:
    _traverser: Traverser
    dom_root: Traversable

    @property
    def traverser(self) -> Traverser:
        if not hasattr(self, "_traverser") or not self._traverser:
            self._traverser = Traverser(self.dom_root)

        return self._traverser

    def elements(
            self,
            id_: Optional[str] = None,
            name: Optional[str] = None,
            class_: Optional[str] = None,
            tag: Optional[str] = None) -> "List[Element]":
        """Search DOM for elements matching provided criteria.

        See :meth:`Traverser.elements` for more documentation.

        :returns: List of Elements matching criteria
        """
        return list(map(
            Element,
            self.traverser.elements(id_, name, class_, tag)))

    def element(
            self,
            id_: Optional[str] = None,
            name: Optional[str] = None,
            class_: Optional[str] = None,
            tag: Optional[str] = None) -> "Element":
        """Search DOM for single element matching provided criteria.

        See :meth:`Traverser.element` for more documentation.

        :returns: Element matching criteria"""
        return Element(self.traverser.element(id_, name, class_, tag))

class Element(LoggerMixin, ElementTraverserMixin):
    """Wrapper for selenium WebElement."""
    _element: WebElement

    def __init__(self, element: WebElement):
        self._element = element

    @property
    def dom_root(self):
        return self._element

    @property
    def tag(self) -> str:
        return cast(str, self._element.tag_name)

    @property
    def id(self) -> Optional[str]:
        return cast(Optional[str], self._element.get_attribute("id"))

    @property
    def class_(self) -> Optional[str]:
        return cast(Optional[str], self._element.get_attribute("class"))

    @property
    def name(self) -> Optional[str]:
        return cast(Optional[str], self._element.get_attribute("name"))

    @property
    def text(self) -> Optional[str]:
        return cast(Optional[str], self._element.text)

    @property
    def selected(self) -> bool:
        return bool(self._element.is_selected())

    @property
    def enabled(self) -> bool:
        return bool(self._element.is_enabled())

    @property
    def displayed(self) -> bool:
        return bool(self._element.is_displayed())

    @property
    def description(self) -> str:
        params = {
            "id": self.id,
            "name": self.name,
            "class": self.class_
        }

        attributes = " ".join(
            [f"{k}=\"{v}\"" for (k, v) in params.items() if v])
        return f"<{self.tag}{attributes and ' '}{attributes}>"

    def __repr__(self):
        return self.description

    def __str__(self):
        return self.description

    def _log_basic_action(self, action: str) -> None:
        self.log.info(f"{action} element {self}")

    def attribute(self, name: str) -> Optional[str]:
        return cast(Optional[str], self._element.get_attribute(name))

    def clear(self) -> None:
        """Clear element's text, if present."""
        self._log_basic_action("clear")
        self._element.clear()

    def click(self) -> None:
        """Click the element."""
        self._log_basic_action("click")
        self._element.click()

    def submit(self) -> None:
        """Submit the element."""
        self._log_basic_action("submit")
        self._element.submit()

    def type(self, text: Optional[str]) -> None:
        """Send text input to the element."""
        self.log.info(f"type '{text}' to element {self}")
        self._element.send_keys(text)

    def screenshot(self, file_path: str) -> None:
        """Save screenshot of element.

        :param file_path: path of file to save screenshot to
        """
        self.log.info(
            f"save screenshot of element {self} to {file_path}")
        self._element.screenshot(file_path)



"""Driver logic and related operations."""

from .driver import Driver
from .element import Element
from .factory import create
from .factory import DriverType
from .factory import DriverFactory

__all__ = [
    "create",
    "Driver",
    "DriverType",
    "DriverFactory",
    "Element"
]



"""Cookie management."""

from typing import cast, Dict, List, Mapping, Optional, Union
import json
import os.path

from ..util import LoggerMixin

COOKIE_FILE = "cookies.json"

def with_valid_expiry(cookie: Mapping) -> Mapping:
    """Convert cookie expiry to milliseconds, if necessary.

    :param cookie: Coookie to be converted
    """
    if "expiry" in cookie and isinstance(cookie["expiry"], float):
        cookie_updated = dict(cookie)
        cookie_updated["expiry"] = int(cookie["expiry"] * 1000)

        return cookie_updated
    else:
        return cookie

class CookieManager(LoggerMixin):
    """Loads, maintains, and persists a list of cookie dicts."""
    _cookies: Optional[List[Mapping]]

    def __init__(self, file_: str = COOKIE_FILE):
        self.cookie_file = file_
        self._cookies = None

    @property
    def cookies(self) -> List[Mapping]:
        """Data structure housing cookies.

        Mapping of domain -> list of associated cookies.
        """
        if self._cookies is None:
            self.cookies = self._load()

        return cast(List[Mapping], self._cookies)

    @cookies.setter
    def cookies(self, cookies: List[Mapping]):
        self._cookies = cookies

    def get(self) -> List[Mapping]:
        """Get stored cookies."""
        return self.cookies

    def add(self, cookies: Union[Dict, List[Dict]]) -> None:
        """Merge cookies into cookie list.

        :param cookies: Cookies to be merged in
        """
        if not isinstance(cookies, list):
            cookies = [cookies]

        new_domains = set(map(lambda c: c["domain"], cookies))
        def not_replaced(cookie: Mapping) -> bool:
            return cookie["domain"] not in new_domains

        new_cookie_list = list(filter(not_replaced, self.cookies))
        new_cookie_list.extend(cookies)

        self.cookies = new_cookie_list

    def _load(self) -> List[Mapping]:
        """Load cookies from cookie file"""
        self.log.info(f"reading saved cookies from {self.cookie_file}")

        if not os.path.exists(self.cookie_file):
            self.log.info(f"cookie file not found: {self.cookie_file}")
            return []

        def is_compatible_cookie(cookie):
            return is_compatible(domain, cookie["domain"])

        with open(self.cookie_file, "r") as file_:
            data = file_.read()
            cookies = json.loads(data)

        return list(map(with_valid_expiry, cookies))

    def save(self, cookies: Optional[Union[Dict, List[Dict]]] = None) -> None:
        """Save cookie list to disk.

        :param cookies: additional cookie(s) to be added before save
        """
        if cookies:
            self.add(cookies)

        self.log.info((
            f"saving {len(self.cookies)} cookies "
            f"to {self.cookie_file}"))

        with open(self.cookie_file, "w") as file_:
            file_.write(json.dumps(self.cookies))

class CookieManagerMixin:
    """Mixin providing CookieManger instance as "cookies" attribute."""
    _cookies: CookieManager

    @property
    def cookies(self) -> CookieManager:
        if not hasattr(self, "_cookies") or not self._cookies:
            self._cookies = CookieManager()

        return self._cookies



"""File utility functions."""

from contextlib import contextmanager
import os
import os.path

@contextmanager
def deleting(file_path: str):
    """Context manager; Silently deletes a specified file on __exit__.

    :param file_path: path of file to be deleted
    """
    try:
        yield
    finally:
        if os.path.exists(file_path):
            os.unlink(file_path)



"""DOM tree traversal functionality."""

from typing import Callable, List, Optional

from selenium.webdriver.remote.webdriver import WebElement

def _attr_filter(attr: str, value: str) -> Callable[[WebElement], bool]:
    """Construct an HTML attribute-based filter predicate.

    Only elements having the provided attribute set to the provided value will
    pass the filter.

    :param attr: name of attribute to filter by
    :param value: value of attr to filter by
    """
    def filter_(element: WebElement) -> bool:
        return bool(element.get_attribute(attr) == value)

    return filter_

def _tag_filter(tag: str) -> Callable[[WebElement], bool]:
    """Construct a HTML tag-based filter predicate.

    Only elements having the provided tag will pass the filter.

    :param tag: tag to filter by
    """
    def filter_(element: WebElement) -> bool:
        return bool(element.tag_name == tag)

    return filter_

def _class_filter(class_: str) -> Callable[[WebElement], bool]:
    """Construct a CSS class-based filter predicate.

    Only elements having the provided class will pass the filter.

    :param class_: name of class to filter by
    """
    def filter_(element: WebElement) -> bool:
        classes = element.get_attribute("class").split(" ")
        return class_ in classes

    return filter_

class Traversable:
    def find_elements_by_id(self, id: str) -> List[WebElement]:
        pass

    def find_elements_by_name(self, name: str) -> List[WebElement]:
        pass

    def find_elements_by_tag_name(self, tag: str) -> List[WebElement]:
        pass

    def find_elements_by_class_name(self, class_: str) -> List[WebElement]:
        pass

    def find_element_by_id(self, id: str) -> WebElement:
        pass

    def find_element_by_name(self, name: str) -> WebElement:
        pass

    def find_element_by_tag_name(self, tag: str) -> WebElement:
        pass

    def find_element_by_class_name(self, class_str) -> WebElement:
        pass

class Traverser:
    """DOM traverser for selenium elements/drivers."""
    dom_root: Traversable

    def __init__(self, dom_root: Traversable):
        self.dom_root = dom_root

    def elements(
            self,
            id_: Optional[str] = None,
            name: Optional[str] = None,
            class_: Optional[str] = None,
            tag: Optional[str] = None) -> List[WebElement]:
        """Search the DOM for elements matching provided criteria.

        At least one criterion must be provided.

        :param id_: HTML id attribute value to filter by
        :param name: HMTL name attribute value to filter by
        :param class_: CSS class to filter by
        :param tag: HTML tag name to filter by
        :returns: List of WebElements matching criteria
        :raises AssertionError: if no criterion is provided
        """
        identifiers_defined = list(filter(bool, [id_, name, class_, tag]))
        assert identifiers_defined, \
            "At least one of id_, name, class_, or tag must be defined"

        filters: List[Callable[[WebElement], bool]] = []
        matches: List[WebElement] = []

        if id_:
            matches.extend(self.dom_root.find_elements_by_id(id_))
            filters.append(_attr_filter("id", id_))

        if name:
            matches.extend(self.dom_root.find_elements_by_name(name))
            filters.append(_attr_filter("name", name))

        if class_:
            matches.extend(self.dom_root.find_elements_by_class_name(class_))
            filters.append(_class_filter(class_))

        if tag:
            matches.extend(self.dom_root.find_elements_by_tag_name(tag))
            filters.append(_tag_filter(tag))

        matches_filtered = matches
        for filter_ in filters:
            matches_filtered = list(filter(filter_, matches))

        return matches_filtered

    def element(
            self,
            id_: Optional[str] = None,
            name: Optional[str] = None,
            class_: Optional[str] = None,
            tag: Optional[str] = None) -> WebElement:
        """Serch the DOM for a single element matching provided criteria.

        Only one criterion should be provided.

        :param id_: HTML id attribute value to filter by
        :param name: HMTL name attribute value to filter by
        :param class_: CSS class to filter by
        :param tag: HTML tag name to filter by
        :returns: WebElement matching criteria or None if no matches were found
        :raises AssertionError: if no or more than one criterion is provided
        :raises selenium.webdriver.common.exceptions.NoSuchElementException: if
            no matching element is found
        """
        identifiers_defined = list(filter(bool, [id_, name, class_, tag]))
        assert len(identifiers_defined) == 1, \
            "Either id_, name, class_, or tag must be defined"

        if id_:
            match = self.dom_root.find_element_by_id(id_)
        elif name:
            match = self.dom_root.find_element_by_name(name)
        elif class_:
            match = self.dom_root.find_element_by_class_name(class_)
        elif tag:
            match = self.dom_root.find_element_by_tag_name(tag)

        return match



__title__ = "aamras"
__description__ = "Library for headless browser manipulation"
__url__ = "https://github.com/allekmott/aamras"
__version__ = "0.0.3"
__author__ = "Allek Mott"
__copyright__ = "2020, Allek Mott"
__author_email__ = "allekmott@gmail.com"
__license__ = "GPLv3"




from .driver import create as create_driver
from .driver import Driver
from .driver import DriverFactory
from .driver import DriverType
from .driver import Element

__all__ = [
    "create_driver",
    "Driver",
    "DriverFactory",
    "DriverType",
    "Element"
]



"""Logging utilities."""

import logging
import sys

class _NiceFormatter(logging.Formatter):
    """winston-esque log formatter"""
    message_format = "%(asctime)s (%(name)s) - %(levelname)s: %(message)s"
    date_format = "%d %b %Y %H:%M:%S"

    def __init__(self):
        super(_NiceFormatter, self).__init__(
            self.message_format, self.date_format)

    def format(self, record):
        if record.levelno < logging.WARNING:
            record.levelname = record.levelname.lower()

        return super(_NiceFormatter, self).format(record)

default_log_formatter = _NiceFormatter()
default_log_handler = logging.StreamHandler(sys.stdout)
default_log_handler.setFormatter(default_log_formatter)
default_level = logging.INFO

_loggers = {}
def get(name=None) -> logging.Logger:
    """Create/get a logger.

    :param name: desired logger name. If None, default logger is returned.
    """
    if name not in _loggers:
        log = logging.getLogger(name) if name else logging.getLogger()

        log.setLevel(default_level)
        if not log.handlers:
            log.addHandler(default_log_handler)

        _loggers[name] = log

    return _loggers[name]

class LoggerMixin:
    """Mixin providing :class:`~logging.Logger` instance as "log" attribute."""
    _logger: logging.Logger

    @property
    def log(self) -> logging.Logger:
        if not hasattr(self, "_logger") or not self._logger:
            self._logger = get(self.__class__.__name__)

        return self._logger



"""Driver wrapper."""

from typing import Optional
import urllib.parse

from selenium.common.exceptions import InvalidCookieDomainException
from selenium.webdriver.remote.webdriver import WebDriver

from ..util import LoggerMixin
from .cookies import CookieManagerMixin
from .element import ElementTraverserMixin

class Driver(LoggerMixin, CookieManagerMixin, ElementTraverserMixin):
    """Abstraction/wrapper of selenium WebDriver."""
    driver: WebDriver

    def __init__(self, driver: WebDriver):
        self.driver = driver

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        self.close()

    @property
    def title(self):
        """Title of the current page."""
        return self.driver.title

    @property
    def url(self):
        """Current URL."""
        return self.driver.current_url

    @property
    def dom_root(self):
        return self.driver

    def close(self):
        """Save cookies and shut down driver."""
        self.log.info("driver shutdown initiated")
        self._save_cookies()

        self.log.info("cleaning up dependencies")
        self.driver.quit()

    def _load_cookies(self) -> None:
        """Load stored cookies and attempt to set them in the driver"""
        self.log.info("loading cookies")
        cookies = self.cookies.get()

        for cookie in cookies:
            try:
                self.driver.add_cookie(cookie)
            except InvalidCookieDomainException:
                pass

    def _save_cookies(self):
        """Save cookies, adding any new cookies from the driver"""
        cookies = self.driver.get_cookies()
        self.cookies.save(cookies)

    def get(self, url: str) -> None:
        """Navigate to a provided url.

        :param url: relative or absolute path, or full URL to navigate to
        """
        url_new = urllib.parse.urljoin(self.driver.current_url, url)

        self.log.info("navigate to %s" % (url_new))
        self.driver.get(url_new)

        self._load_cookies()
        self.log_page()

    def click(
            self,
            id_: Optional[str] = None,
            name: Optional[str] = None,
            class_: Optional[str] = None,
            tag: Optional[str] = None) -> None:
        """Click an element matching provided criteria.

        See :meth:`element` for documentation on criteria.
        """
        element = self.element(id_, name, class_, tag)
        element.click()

        self.log_page()

    def submit(
            self,
            id_: Optional[str] = None,
            name: Optional[str] = None,
            class_: Optional[str] = None,
            tag: Optional[str] = None) -> None:
        """Submit an element matching provided criteria.

        See :meth:`element` for documentation on criteria.
        """
        element = self.element(id_, name, class_, tag)
        element.click()

        self.log_page()

    def type(
            self,
            id_: Optional[str] = None,
            name: Optional[str] = None,
            class_: Optional[str] = None,
            text: Optional[str] = None) -> None:
        """Send text input to an element matching provided criteria.

        See :meth:`element` for documentation on criteria.
        """
        element = self.element(id_, name, class_)
        element.type(text)

    def screenshot(self, file_path: str):
        """Save a screenshot of the browser window.

        :param file_path: path of file to save screenshot to
        """
        self.log.info("saving screenshot to '%s'" % (file_path))
        self.driver.get_screenshot_as_file(file_path)

    def save_source(self, file_path: str) -> None:
        """Save source code of current page.

        :param file_path: path of file to save source to
        """
        self.log.info(f"write page source to {file_path}")
        with open(file_path, "w") as file_:
            file_.write(str(self.driver.page_source))

    def log_page(self) -> None:
        """Write page information to log."""
        self.log.info(
            f"page title is {self.driver.title} ({self.driver.current_url})")



import os.path
from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

path_readme = os.path.join(here, "README.rst")
path_version = os.path.join(here, "aamras", "__version__.py")

with open(path_readme) as file_:
    long_description = file_.read()

about = {}
with open(path_version) as file_:
    exec(file_.read(), about)

setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    project_urls={
        "Documentation": "https://aamras.readthedocs.io/",
        "Source": about["__url__"],
    },
    packages=find_packages(exclude=["tests", "tests.*"]),
    license=about["__license__"],
    data_files=[("", ["LICENSE"])],
    python_requires=">=3.6",
    install_requires=["selenium"],
    extras_require={
        "dev": ["flake8", "mypy", "pytest", "pytest-cov", "pipenv-setup", "sphinx",]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries",
    ],
)
